import os
import sys
import subprocess
from pkg_resources import resource_filename

def get_instance_from_path(path: str) -> str:
	
	roblox_path_keys = path.split("/")

	roblox_path = ""

	for i, key in enumerate(roblox_path_keys):
		if i == 0:
			roblox_path += key
		else:
			roblox_path += f":WaitForChild(\"{key}\")"

	for service in ["ReplicatedStorage", "ServerStorage", "ServerScriptService", "ReplicatedFirst", "Lighting", "StarterGui", "StarterPlayer", "Workspace"]:
		roblox_path = roblox_path.replace(f"game:WaitForChild(\"{service}\")", f"game:GetService(\"{service}\")")

	return roblox_path

def run_bundled_exe(exe_name, args=None):
	# Get the path to the bundled executable file within the package
	exe_file_path = resource_filename('luau', f'data/{exe_name}')

	# Check if running as a PyInstaller single-file executable
	if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
		# Construct the path to the executable file in a PyInstaller bundle
		exe_file_path = os.path.join(sys._MEIPASS, 'data', exe_name)

	# Ensure the file is executable
	os.chmod(exe_file_path, 0o755)

	# Build the command to run, including the executable path and any arguments
	cmd = [exe_file_path]
	if args:
		cmd.extend(args)

	# Run the bundled executable with the specified arguments
	subprocess.run(cmd, check=True)

def get_module_require(path: str):
	return f"require({get_instance_from_path(path)})"

def get_header_module(path: str, variable_name: str = ""):
	if variable_name == "":
		keys = path.split("/")
		variable_name = keys[len(keys)-1]

	value = get_module_require(path)

	return f"local {variable_name} = {value}"



# creates a recursive luau table
def import_type(module_variable_name: str, type_name: str, type_local_name="") -> str:
	if type_local_name == "":
		type_local_name = type_name

	return f"type {type_local_name} = {module_variable_name}.{type_name}"	

def indent_block(content: list[str], indent_count: int = 1) -> list[str]:
	out = []
	for line in content:
		out.append(("\t"*indent_count)+line)

	return out

def define_variable(
	variable_name: str, 
	value: str | None = None, 
	type_name: str | None = None, 
	is_private: bool = False, 
	include_semi_colon: bool = False
):
	final_variable_name = variable_name
	if is_private:
		final_variable_name = "_"+final_variable_name

	definition = f"local {final_variable_name}"
	if type_name != None:
		definition += f": {type_name}"

	statement = definition
	if value != None:
		statement = f"{definition} = {value}"

	if include_semi_colon:
		return statement + ";"
	else:
		return statement	

def get_function_header(name: str = "", parameters: list=[], return_type="", is_local: bool=False) -> str:
	assert name != "" or is_local==False, "you can't have a nameless local function"

	final_name_text = ""
	if name != "":
		final_name_text = " "+name

	local_text = ""
	if is_local:
		local_text = "local "

	param_text = ",".join(parameters)

	return_text = ""
	if return_type != "":
		return_text += f": {return_type}"

	return f"{local_text}function{final_name_text}({param_text}){return_text}"

def get_function_call(name: str="", object_name:str="", parameters: list=[]) -> str:
	object_text = ""
	if object_name != "":

		def get_first_letter(string):
			if not string:
				return None
			for char in string:
				if char.isalpha():
					return char
			
			return None

		first_letter = get_first_letter(name)
		if first_letter != None:
			if first_letter.isupper():
				object_text = object_text+":"
			else:
				object_text = object_text+"."
		else:
			object_text = object_name

	reference = object_text+name

	param_text = ",".join(parameters)

	return f"{reference}({param_text})"



import os
import dpath
import json
from .util import run_bundled_exe

DEFAULT_ROJO_PROJECT_PATH = "default.project.json"
ROJO_SOURCE = "rojo-rbx/rojo"
ROJO_VERSION = "7.1.0"

def get_rojo_project_path() -> str:
	if os.path.exists(DEFAULT_ROJO_PROJECT_PATH):
		return DEFAULT_ROJO_PROJECT_PATH
	for file_path in os.listdir(""):
		if os.path.isfile(file_path):
			base, ext = os.path.splitext(file_path)
			if ext == ".json":
				sec_base, sec_ext = os.path.splitext(base)
				if sec_ext== ".project":
					return file_path

def build_sourcemap(project_json_path: str = ""):
	if project_json_path == "":
		project_json_path = get_rojo_project_path()
	run_bundled_exe(exe_name="rojo.exe", args=["sourcemap", project_json_path, "--output", "sourcemap.json"])

def get_roblox_path_from_env_path(env_path: str) -> str:
	project_path = get_rojo_project_path()
	rojo_file = open(project_path, "r")
	rojo_config = json.loads(rojo_file.read())
	rojo_file.close()
	tree_config = rojo_config["tree"]
	options = {}
	env_length = len(env_path.split("/"))
	best_env_path = ""
	best_env_length = 0
	for tree_path, tree_val in dpath.search(tree_config, '**', yielded=True):
		keys = tree_path.split("/")
		final_key = keys[len(keys)-1]
		if final_key == "$path" and len(keys) > 1:
			ro_path = "game/" + "/".join(keys[0:(len(keys)-1)])
			options[tree_val] = ro_path
			val_length = len(tree_val.split("/"))
			if val_length <= env_length and best_env_length < val_length:
				pattern = env_path[0:len(tree_val)]
				if pattern == tree_val:
					best_env_path = tree_val
					best_env_length = val_length

	conclusion = env_path[(len(best_env_path)):]
	final = options[best_env_path] + conclusion

	return final.split(".")[0]




import os
import toml

FOREMAN_DEFAULT_PATH = "foreman.toml"
AFTMAN_DEFAULT_PATH = "aftman.toml"

def get_tool_name(source: str, version: str) -> str:
	alt_nickname = source.split("/")[len(source.split("/"))-1]
	
	version_str = version.replace(".","_").lower()
	if version_str[0] == "v":
		version_str = version_str[1:]

	version_str = "_VERSION_"+version_str

	if os.path.exists(FOREMAN_DEFAULT_PATH):
		foreman_config = toml.loads(open(FOREMAN_DEFAULT_PATH, "r").read())
		foreman_tools = foreman_config["tools"]


		for nickname, entry in foreman_tools.items():
			if nickname == alt_nickname:
				alt_nickname = alt_nickname+version_str

			if "source" in entry and "version" in entry:
				if entry["source"] == source and entry["version"] == version:
					return nickname

		foreman_tools[alt_nickname] = {
			"source": source,
			"version": version
		}

		foreman_file = open(FOREMAN_DEFAULT_PATH, "w")
		foreman_file.write(toml.dumps(foreman_config))
		foreman_file.close()

		os.system("foreman install")

		return alt_nickname

	elif os.path.exists(AFTMAN_DEFAULT_PATH):
		aftman_config = toml.loads(open(AFTMAN_DEFAULT_PATH, "r").read())
		aftman_tools = aftman_config["tools"]
		aft_path = source+"@"+version
		for nickname, path in aftman_tools.items():
			if nickname == alt_nickname:
				alt_nickname = alt_nickname+version_str
			if path == aft_path:
				return path
		
		os.system(f"aftman add {aft_path} -{alt_nickname}")
		return alt_nickname
	else:
		raise IndexError("no foreman or aftman config files found")



import os
from .util import run_bundled_exe

def format(path: str):
	# stylua_tool_name = get_tool_name("JohnnyMorganz/StyLua", "0.15.3")
	# os.system(f"{stylua_tool_name} {path}")
	run_bundled_exe(exe_name="stylua.exe", args=[path])

def write_script(build_path: str, content: str, write_as_directory: bool=False):
	dir_name, file_name = os.path.split(build_path)
	if not os.path.exists(dir_name):
		os.makedirs(dir_name)
	elif os.path.exists(build_path):
		os.remove(build_path)

	if write_as_directory:
		full_dir_path = build_path.split(".")[0]
		full_ext = ".".join(build_path.split(".")[1:])
		final_dir_path = (dir_name+"/"+file_name).replace(full_ext, "")
		os.makedirs(final_dir_path)
		init_file_path = final_dir_path+"/init."+full_ext
		out_file = open(init_file_path, "w")
		out_file.write(content)
		out_file.close()
		format(init_file_path)

	else:
		out_file = open(build_path, "w")
		out_file.write(content)
		out_file.close()
		format(build_path)




import os
import toml
from .util import run_bundled_exe
from .tool import get_tool_name
from .rojo import get_rojo_project_path, build_sourcemap

WALLY_SOURCE = "UpliftGames/wally"
WALLY_VERSION = "0.3.1"
WPT_SOURCE = "JohnnyMorganz/wally-package-types"
WPT_VERSION = "1.2.0"

def get_wally_name():
	return get_tool_name(WALLY_SOURCE, WALLY_VERSION)

def update_wally():

	project_path = get_rojo_project_path()
	wally_tool_name = get_wally_name()

	os.system(f"{wally_tool_name} install")
	build_sourcemap()
	run_bundled_exe(exe_name="wally-package-types.exe", args=["--sourcemap", "sourcemap.json", "Packages"])

def get_wally_package_nickname(package_wally_path: str) -> str:
	wally_config = toml.loads(open("wally.toml", "r").read())
	generated_nickname = (package_wally_path.split("/")[1].split("@")[0]).title()
	name_parts = generated_nickname.split("-")
	for i, name_part in enumerate(name_parts):
		name_parts[i] = name_part.title()

	generated_nickname = "".join(name_parts)

	version = package_wally_path.split("@")[1]
	version_str = version.replace(".","_").lower()
	if version_str[0] == "v":
		version_str = version_str[1:]

	version_str = "_VERSION_"+version_str

	for nickname, package_path in wally_config["dependencies"].items():
		if nickname == generated_nickname:
			generated_nickname = generated_nickname + version_str
		if package_wally_path == package_path:
			return nickname

	# no maid installed
	wally_config["dependencies"][generated_nickname] = package_wally_path

	open("wally.toml", "w").write(toml.dumps(wally_config))
	update_wally()

	return generated_nickname

def require_roblox_wally_package(
	package_wally_path: str, 
	is_private: bool = False, 
	is_header: bool = True,
	wally_directory_path: str ="game/ReplicatedStorage/Packages"
) -> str:

	package_nickname = get_wally_package_nickname(package_wally_path)

	package_path = ""
	roblox_path_keys = (wally_directory_path+"/"+package_nickname).split("/")
	for i, key in enumerate(roblox_path_keys):
		if i == 0:
			package_path += key
		else:
			package_path += f":WaitForChild(\"{key}\")"

	package_path = package_path.replace("game:WaitForChild(\"ReplicatedStorage\")", "game:GetService(\"ReplicatedStorage\")")
	package_path = package_path.replace("game:WaitForChild(\"ServerStorage\")", "game:GetService(\"ServerStorage\")")
	package_path = package_path.replace("game:WaitForChild(\"ReplicatedFirst\")", "game:GetService(\"ReplicatedFirst\")")

	if is_header:
		if is_private:
			return f"local _{package_nickname} = require({package_path})"
		else:
			return f"local {package_nickname} = require({package_path})"
	else:
		return f"require({package_path})"




import dpath

READ_STR_AS_LITERAL_PREFIX = "<LITERAL_STRING_START>_!_"
READ_STR_AS_LITERAL_SUFFIX = "_!_<LITERAL_STRING_END>"

def _insert_comma(enabled: bool) -> str:
	if enabled:
		return ","
	else:
		return ""

def _get_indent(indent_count: int) -> str:
	return "\t"*indent_count

def from_bool(value: bool, indent_count=0, add_comma_at_end=False) -> str:
	return _get_indent(indent_count) + str(value).lower() + _insert_comma(add_comma_at_end)

def get_if_literal(value: str):
	return (READ_STR_AS_LITERAL_PREFIX == value[0:len(READ_STR_AS_LITERAL_PREFIX)]) or value.find(READ_STR_AS_LITERAL_SUFFIX) != -1

def from_str(value: str, indent_count=0, add_comma_at_end=False) -> str:
	if get_if_literal(value):
		raw_val = value.replace(READ_STR_AS_LITERAL_PREFIX, "").replace(READ_STR_AS_LITERAL_SUFFIX, "")
		return _get_indent(indent_count) + raw_val + _insert_comma(add_comma_at_end)
	else:
		return _get_indent(indent_count) + f"\"{value}\"" + _insert_comma(add_comma_at_end)

def from_number(value: int | float, indent_count=0, add_comma_at_end=False) -> str:
	return _get_indent(indent_count) + str(value) + _insert_comma(add_comma_at_end)

def from_nil(indent_count=0, add_comma_at_end=False) -> str:
	return f"{_get_indent(indent_count)} nil" + _insert_comma(add_comma_at_end)

def from_list(value: list, indent_count=0, add_comma_at_end=False, multi_line=True, skip_initial_indent=False):
	
	# start list
	list_val = ""
	if skip_initial_indent:
		list_val += "{"
	else:
		list_val += _get_indent(indent_count) + "{"

	# iterate through values
	for v in value:
		
		# write entry
		if type(v) == dict or type(v) == list:
			entry = from_any(v, indent_count+1, False, multi_line, True) + _insert_comma(True)
		else:
			entry = from_any(v, indent_count, False, multi_line, True) + _insert_comma(True)

		# add it to existing string
		if multi_line:
			list_val += "\n" + _get_indent(indent_count+1) + entry
		else:
			list_val += entry

	# end the table on a new line if multi-line
	if multi_line:
		list_val += "\n" + _get_indent(indent_count)

	# close the table
	list_val += "}"

	# indent as needed and return value
	return _get_indent(indent_count) + list_val + _insert_comma(add_comma_at_end)

def from_dict(value: dict, indent_count=0, add_comma_at_end=False, multi_line=True, skip_initial_indent=False):
	
	# start dictionary
	list_val = ""
	if skip_initial_indent:
		list_val += "{"
	else:
		list_val += _get_indent(indent_count) + "{"

	# iterate through key-val pairs
	for k, v in value.items():

		is_literal = get_if_literal(k)

		# write entry
		if type(v) == dict or type(v) == list:
			if is_literal:
				entry = f"{from_any(k, 0, False)} = {from_any(v, indent_count+1, False, multi_line, True)}" + _insert_comma(True)
			else:
				entry = f"[{from_any(k, 0, False)}] = {from_any(v, indent_count+1, False, multi_line, True)}" + _insert_comma(True)
		else:
			if is_literal:
				entry = f"{from_any(k, 0, False)} = {from_any(v, 0, False, multi_line, True)}" + _insert_comma(True)
			else:
				entry = f"[{from_any(k, 0, False)}] = {from_any(v, 0, False, multi_line, True)}" + _insert_comma(True)

		# add it to existing string
		if multi_line:
			list_val += "\n" + _get_indent(indent_count+1) + entry
		else:
			list_val += entry

	# end the table on a new line if multi-line
	if multi_line:
		list_val += "\n" + _get_indent(indent_count)

	# close the table
	list_val += "}"

	# indent as needed and return value
	if skip_initial_indent:
		return _get_indent(0) + list_val + _insert_comma(add_comma_at_end)
	else:
		return _get_indent(indent_count) + list_val + _insert_comma(add_comma_at_end)

def from_dict_to_type(type_value: dict, indent_count=0, add_comma_at_end=False, multi_line=True, skip_initial_indent=False) -> str:
	out = {}
	for path, value in dpath.search(type_value, '**', yielded=True):
		literal_path_keys = []
		for key in path.split("/"):
			literal_path_keys.append(mark_as_literal(key))

		literal_path = "/".join(literal_path_keys)

		if value == bool or type(value) == bool:
			dpath.new(out, literal_path, mark_as_literal("boolean"))
		elif value == int or value == float or type(value) == int or type(value) == float:
			dpath.new(out, literal_path, mark_as_literal("number"))
		elif value == str:
			dpath.new(out, literal_path, mark_as_literal("string"))	
		elif type(value) == str:
			if get_if_literal(value):
				dpath.new(out, literal_path, value)
			else:
				dpath.new(out, literal_path, mark_as_literal("string"))	

	return from_dict(out, indent_count, add_comma_at_end, multi_line, skip_initial_indent).replace(" = ", ": ")

def from_any(
	value: int | str | None | float | dict | list = None, 
	indent_count = 0, 
	add_comma_at_end = False, 
	multi_line=True,
	skip_initial_indent=False
) -> str:

	if type(value) == list:

		return from_list(value, indent_count, add_comma_at_end, multi_line, skip_initial_indent)

	elif type(value) == dict:
		
		return from_dict(value, indent_count, add_comma_at_end, multi_line, skip_initial_indent)

	elif type(value) == float or type(value) == int:
		return from_number(value, indent_count, add_comma_at_end)
		
	elif type(value) == bool:
		return from_bool(value, indent_count, add_comma_at_end)
		
	elif type(value) == str:
		return from_str(value, indent_count, add_comma_at_end)
		
	return from_nil(indent_count, add_comma_at_end)


# allows for processing of text in the other methods as is without being wrapped in quotes as a string
def mark_as_literal(text: str) -> str:
	return READ_STR_AS_LITERAL_PREFIX+text+READ_STR_AS_LITERAL_SUFFIX